/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - GUI
 *  Copyright (C) 2022 Tibor 'Igor2' Palinkas
 *
 *  (Supported by NLnet NGI0 PET Fund in 2022)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */

/* text object dialogs */

#include <libcschem/config.h>

#include <string.h>
#include <genht/htsp.h>
#include <genht/hash.h>

#include <librnd/core/actions.h>
#include <librnd/core/hidlib.h>
#include <librnd/core/hid_dad.h>

#include <libcschem/concrete.h>
#include <libcschem/cnc_text.h>

#include <sch-rnd/search.h>
#include <sch-rnd/operation.h>
#include <sch-rnd/crosshair.h>
#include <sch-rnd/draw.h>
#include <sch-rnd/funchash_core.h>

#include "timed_chg.h"

typedef struct {
	RND_DAD_DECL_NOINIT(dlg)
	csch_sheet_t *sheet;
	csch_text_t *text;
	int wsubst1, wtext;
	int start, len; /* character offset of start and length of the first %% substitution */
	int gui_active;

	struct {
		csch_timed_chg_t hdr;
		gds_t str;
	} timed;

	gds_t tmp;
} dyntext_dlg_ctx_t;

static void dyntext_dlg_text2dlg(dyntext_dlg_ctx_t *ctx)
{
	char *first = "", *last;

	ctx->start = ctx->len = -1;
	ctx->tmp.used = 0;
	gds_append_str(&ctx->tmp, ctx->text->text);

	if (ctx->tmp.used > 2) {
		first = strchr(ctx->tmp.array, '%');
		if (first != NULL) {
			first++;
			last = strchr(first, '%');
			if ((last != NULL) && (last > first+1)) {
				*last = '\0';
				ctx->start = first - ctx->tmp.array;
				ctx->len = last - first;
			}
			else
				first = "";
		}
		else
			first = "";
	}

	RND_DAD_SET_VALUE(ctx->dlg_hid_ctx, ctx->wsubst1, str, first);
	RND_DAD_SET_VALUE(ctx->dlg_hid_ctx, ctx->wtext, str, ctx->text->text);
}

static void set_text_timed(void *ctx_)
{
	dyntext_dlg_ctx_t *ctx = ctx_;

	csch_op_text_edit(ctx->sheet, ctx->text, ctx->timed.str.array);
	rnd_gui->invalidate_all(rnd_gui);
	if (ctx->gui_active)
		dyntext_dlg_text2dlg(ctx);
	ctx->timed.str.used = 0;
}

csch_inline void dyntext_build_timed_subst(dyntext_dlg_ctx_t *ctx, const char *str1, const char *str2)
{
	ctx->timed.str.used = 0;
	gds_append_len(&ctx->timed.str, ctx->text->text, ctx->start);
	if (str1 != NULL)
		gds_append_str(&ctx->timed.str, str1);
	if (str2 != NULL)
		gds_append_str(&ctx->timed.str, str2);
	gds_append_str(&ctx->timed.str, ctx->text->text + ctx->start + ctx->len);
}

static void dyntext_subst1_cb(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	dyntext_dlg_ctx_t *ctx = caller_data;

	if (ctx->start < 0)
		return;

	dyntext_build_timed_subst(ctx, ctx->dlg[ctx->wsubst1].val.str, NULL);
	csch_timed_chg_schedule(&ctx->timed.hdr);
}

static void dyntext_raw_cb(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	dyntext_dlg_ctx_t *ctx = caller_data;

	ctx->timed.str.used = 0;
	gds_append_str(&ctx->timed.str, ctx->dlg[ctx->wtext].val.str);
	csch_timed_chg_schedule(&ctx->timed.hdr);
}

static void dyntext_pick_cb(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr)
{
	dyntext_dlg_ctx_t *ctx = caller_data;
	fgw_arg_t res, args[2];
	csch_oidpath_t idp = {0};
	gds_t op = {0};
	const char *key = NULL;

	/* prepare first arg */
	csch_oidpath_from_obj(&idp, &ctx->text->hdr.parent->hdr);
	gds_append_str(&op, "object:");
	csch_oidpath_to_str_append(&op, &idp);
	csch_oidpath_free(&idp);

	/* call action, save resulting attr key */
	args[1].type = FGW_STR;
	args[1].val.cstr = op.array;
	if (rnd_actionv_bin(&ctx->sheet->hidlib, "attributepick", &res, 2, args) == 0)
		key = res.val.cstr;

	/* clean up action call */
	fgw_arg_free(&rnd_fgw, &res);
	gds_uninit(&op);

	if ((key == NULL) || (*key == '\0'))
		return; /* cancel */

	/* build and update */
	dyntext_build_timed_subst(ctx, "../A.", key);
	csch_op_text_edit(ctx->sheet, ctx->text, ctx->timed.str.array);
	rnd_gui->invalidate_all(rnd_gui);
	dyntext_dlg_text2dlg(ctx);
	ctx->timed.str.used = 0;
}


static int sch_rnd_edit_text_dialog_dynamic(csch_sheet_t *sheet, csch_text_t *text)
{
	static const char help_subst1[] = "Quick edit for the first %% substitution";
	static const char help_raw[] = "Edit text string as-is; presents the template without any transformation";
	static const char spaces[] = "                            ";
	rnd_hid_dad_buttons_t clbtn[] = {{"Close", 0}, {NULL, 0}};
	dyntext_dlg_ctx_t ctx = {0};

	ctx.sheet = sheet;
	ctx.text = text;

	csch_timed_chg_init(&ctx.timed.hdr, set_text_timed, &ctx);

	RND_DAD_BEGIN_VBOX(ctx.dlg);
		RND_DAD_COMPFLAG(ctx.dlg, RND_HATF_EXPFILL);

		RND_DAD_BEGIN_HBOX(ctx.dlg);
			RND_DAD_BEGIN_VBOX(ctx.dlg);
				RND_DAD_LABEL(ctx.dlg, spaces);
				RND_DAD_LABEL(ctx.dlg, "First subst:");
					RND_DAD_HELP(ctx.dlg, help_subst1);
				RND_DAD_LABEL(ctx.dlg, spaces);
			RND_DAD_END(ctx.dlg);

			RND_DAD_STRING(ctx.dlg);
				RND_DAD_HELP(ctx.dlg, help_subst1);
				ctx.wsubst1 = RND_DAD_CURRENT(ctx.dlg);
				RND_DAD_CHANGE_CB(ctx.dlg, dyntext_subst1_cb);

			RND_DAD_BEGIN_VBOX(ctx.dlg);
				RND_DAD_LABEL(ctx.dlg, "");
				RND_DAD_BUTTON(ctx.dlg, "Pick...");
					RND_DAD_HELP(ctx.dlg, "Pick an existing attribute from the parent group");
					RND_DAD_CHANGE_CB(ctx.dlg, dyntext_pick_cb);
				RND_DAD_LABEL(ctx.dlg, "");
			RND_DAD_END(ctx.dlg);

		RND_DAD_END(ctx.dlg);

		RND_DAD_BEGIN_HBOX(ctx.dlg);
			RND_DAD_BEGIN_VBOX(ctx.dlg);
				RND_DAD_LABEL(ctx.dlg, spaces);
				RND_DAD_LABEL(ctx.dlg, "Raw (template):");
					RND_DAD_HELP(ctx.dlg, help_raw);
				RND_DAD_LABEL(ctx.dlg, spaces);
			RND_DAD_END(ctx.dlg);
			RND_DAD_STRING(ctx.dlg);
				RND_DAD_WIDTH_CHR(ctx.dlg, 32);
				RND_DAD_HELP(ctx.dlg, help_raw);
				ctx.wtext = RND_DAD_CURRENT(ctx.dlg);
				RND_DAD_CHANGE_CB(ctx.dlg, dyntext_raw_cb);
		RND_DAD_END(ctx.dlg);

		CSCH_DAD_TIMING(ctx.dlg, &ctx.timed.hdr, "text");

		RND_DAD_BUTTON_CLOSES(ctx.dlg, clbtn);
	RND_DAD_END(ctx.dlg);

	RND_DAD_NEW("DyntextDialog", ctx.dlg, "Dynamic text edit", &ctx, 1, NULL);

	csch_timed_chg_timing_init(&ctx.timed.hdr, ctx.dlg_hid_ctx);
	dyntext_dlg_text2dlg(&ctx);

	ctx.gui_active = 1;
	RND_DAD_RUN(ctx.dlg);
	ctx.gui_active = 0;
	ctx.timed.hdr.wtiming = -1;

	csch_timed_chg_finalize(&ctx.timed.hdr);

	RND_DAD_FREE(ctx.dlg);

	gds_uninit(&ctx.tmp);
	gds_uninit(&ctx.timed.str);
	return 0;
}


static int sch_rnd_edit_text_dialog_static(csch_sheet_t *sheet, csch_text_t *text)
{
	char *new_str = rnd_hid_prompt_for(&sheet->hidlib, "Edit text object:", text->text, "Edit text object");

	if (new_str != NULL) {
		csch_op_text_edit(sheet, text, new_str);
		rnd_gui->invalidate_all(rnd_gui);
		free(new_str);
		return 0;
	}
	return -1;
}


int sch_rnd_edit_text_dialog(csch_sheet_t *sheet, csch_text_t *text)
{
	return text->dyntext ? sch_rnd_edit_text_dialog_dynamic(sheet, text) : sch_rnd_edit_text_dialog_static(sheet, text);
}


const char csch_acts_EditText[] = "EditText([object[=idpath]])";
const char csch_acth_EditText[] = "Bring up a text quick edit dialog.\n";
fgw_error_t csch_act_EditText(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_hidlib_t *hidlib = RND_ACT_HIDLIB;
	csch_sheet_t *sheet = (csch_sheet_t *)hidlib;
	int cmd;

	TODO("scripting: pick up idpath and set grp");
	RND_ACT_CONVARG(1, FGW_KEYWORD, EditText, cmd = fgw_keyword(&argv[1]));

	RND_ACT_IRES(-1);

	switch(cmd) {
		case F_Object:
			{
				csch_chdr_t *obj;
				csch_coord_t x, y;

				if (csch_get_coords("Click on text to edit", &x, &y, 0) != 0)
					break;

				obj = csch_search_obj_at(sheet, x, y, csch_slop);
				if ((obj != NULL) && (obj->type == CSCH_CTYPE_TEXT)) {
					int r = sch_rnd_edit_text_dialog(sheet, (csch_text_t *)obj);
					RND_ACT_IRES(r);
				}
				else
					rnd_message(RND_MSG_ERROR, "EditText(): no text object under cursor\n");
			}
			break;
		default:
			rnd_message(RND_MSG_ERROR, "EditText(): invalid first argument\n");
	}

	return 0;
}

